"""
Common classes for TaskStore implementations, including the base TaskStore class.

"""
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Iterable, Optional

from kombu import Message


@dataclass
class StoredTask:
    """
    A Celery task.
    """

    headers: Dict[str, Any]
    body: Any
    routing_key: str
    # args: List[Any]
    # kwargs: Dict[str, Any]

    def json(self, indent=2):
        """
        Serialize to JSON.
        """
        return json.dumps(
            {
                "headers": self.headers,
                "body": self.body,
                "routing_key": self.routing_key,
            },
            indent=indent,
        )

    @classmethod
    def from_string(cls, string):
        """
        Instantiate a StoredTask from a JSON string.
        """
        data = json.loads(string)
        return cls(**data)

    @classmethod
    def from_message(cls, message: Message):
        """
        Instantiate a StoredTask from a kombu Message.
        """
        # Kombu docs say message.body is a str, but it's really a memoryview,
        # which is not JSON serializable
        return cls(
            body=message.decode(),
            headers=message.headers,
            routing_key=message.delivery_info["routing_key"],
            # args=message.payload[0],
            # kwargs=message.payload[1],
        )

    @property
    def id(self):
        """
        The task ID assigned by Celery.
        """
        return self.headers["id"]

    @property
    def task(self):
        """
        Python path of the task function.
        """
        return self.headers["task"]

    @property
    def argsrepr(self):
        """
        Representation of the positional arguments
        which will be passed to the task.
        """
        return self.headers["argsrepr"]

    @property
    def kwargsrepr(self):
        """
        Representation of the keyword arguments
        which will be passed to the task.
        """
        return self.headers["kwargsrepr"]

    def __repr__(self):
        return f"<StoredTask {self.task}: {self.id}>"


class TaskStore(ABC):
    """
    Abstract base class for storing tasks.
    """

    @abstractmethod
    def save(self, task: StoredTask):
        """
        Save the task to some persistance layer.
        Subclasses must implement this.
        """
        ...

    def bulk_save(self, tasks: Iterable[StoredTask]):
        """
        Save multiple tasks.
        """
        for task in tasks:
            self.save(task)

    @abstractmethod
    def load_tasks(self, task_name=Optional[str]) -> Iterable[StoredTask]:
        """
        Load tasks from some persistence layer.
        Subclasses must implement this.
        """
        ...

    @abstractmethod
    def delete(self, task: StoredTask):
        """
        Remove a task from the persistence layer.
        Subclasses must implement this.
        """
        ...

    # Optional functionality

    def dedupe(self) -> int:
        """
        Remove duplicate tasks from the store.

        Returns the number of tasks removed.
        """
        raise NotImplementedError()



import os
from pathlib import Path
from typing import Iterable, Optional

from taskrabbit.config import FileConfig
from .base import TaskStore, StoredTask


class FileTaskStore(TaskStore):
    config_class = FileConfig

    def __init__(self, cfg: FileConfig):
        self.path = Path() / cfg.directory
        self.path.mkdir(parents=True, exist_ok=True)

    def save(self, task: StoredTask):
        with open(self.path / task.id, "w") as f:
            f.write(task.json())

    def load_tasks(self, task_name: Optional[str] = None) -> Iterable[StoredTask]:
        for path in self.path.glob("*"):
            with open(path) as f:
                data = f.read()
                task = StoredTask.from_string(data)
                if task_name is None or task.task == task_name:
                    yield task

    def delete(self, task: StoredTask):
        try:
            os.remove(self.path / task.id)
        except FileNotFoundError:
            pass



import typer
from contextlib import contextmanager
from importlib import import_module

from halo import Halo

SPINNER = "dots12"


def pluralize(count: int):
    return "" if count == 1 else "s"


def green(obj):
    return typer.style(str(obj), fg=typer.colors.GREEN)


def red(obj):
    return typer.style(str(obj), fg=typer.colors.RED)


# From django.utils.module_loading
def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class'
            % (module_path, class_name)
        ) from err


@contextmanager
def spinner(text: str):
    with Halo(text=text, spinner=SPINNER):
        yield



import configparser

from dataclasses import dataclass
from pathlib import Path
from typing import Mapping

from taskrabbit.utils import import_string
from taskrabbit.stores.base import TaskStore


DEFAULT_LOG_LEVEL = "INFO"

# Sets kombu Consumer prefetch count
# See https://docs.celeryproject.org/projects/kombu/en/stable/userguide/consumers.html#reference  # noqa
DEFAULT_CONSUMER_PREFETCH_COUNT = 100


DEFAULTS = {
    "taskrabbit": {"store": "sqlite", "log_level": "INFO"},
    "rabbitmq": {
        "username": "guest",
        "password": "guest",
        "host": "localhost",
        "port": "5672",
        "vhost": "/",
    },
    "sqlite": {"db": "tasks.sqlite"},
}

CONFIG_PATHS = [Path.home() / ".taskrabbit.ini", Path("taskrabbit.ini")]


class ConfigurationError(Exception):
    pass


@dataclass(frozen=True)
class RabbitMQConfig:
    username: str
    password: str
    host: str
    port: str = "5672"
    vhost: str = "/"
    consumer_prefetch_count: int = DEFAULT_CONSUMER_PREFETCH_COUNT

    def url(self):
        return (
            f"amqp://{self.username}:{self.password}@"
            f"{self.host}:{self.port}{self.vhost}"
        )

    def __post_init__(self):
        if not self.vhost.startswith("/"):
            raise ValueError(f"{self.__class__.__name__}.vhost must have a leading /")


class StoreConfig:
    name: str


@dataclass
class Config:
    store_config: StoreConfig
    log_level: str
    rabbitmq: RabbitMQConfig
    store_class: TaskStore

    @classmethod
    def from_config_dict(cls, cfg: Mapping):
        store_path = cfg["taskrabbit"]["store"]
        del cfg["taskrabbit"]["store"]
        store_cls = import_string(store_path)
        store_config_cls = store_cls.config_class
        if "store" in cfg:
            store_cfg = store_config_cls(**cfg["store"])
        else:
            store_cfg = store_config_cls()
        rabbit_cfg = RabbitMQConfig(**cfg["rabbitmq"])
        config = cls(
            rabbitmq=rabbit_cfg,
            store_config=store_cfg,
            store_class=store_cls,
            **cfg["taskrabbit"],
        )
        return config

    def init_store(self):
        return self.store_class(self.store_config)


@dataclass(frozen=True)
class SqliteConfig(StoreConfig):
    db: str = "tasks.sqlite"
    name: str = "sqlite"


@dataclass(frozen=True)
class PostgresConfig(StoreConfig):
    username: str
    password: str
    host: str
    port: str
    db: str
    name: str = "postgres"

    def get_dsn(self):
        return (
            f"postgresql://{self.username}:{self.password}@"
            f"{self.host}:{self.port}/{self.db}"
        )


@dataclass(frozen=True)
class FileConfig(StoreConfig):
    name = "file"
    directory = "tasks"


def _update_config(config, options):
    """
    Recursively overwrite keys in `config` with values from `options`.
    """
    for k, v in config.items():
        if k in options:
            if hasattr(v, "items"):
                _update_config(config[k], options[k])
            else:
                config[k] = options[k]


def merge_config_files_and_options(*paths: Path, **opts) -> configparser.ConfigParser:
    """
    Load + merge multiple config paths, overriding values with
    kwarg options.

    Will raise ConfigurationError if at least one path is not found.
    """
    config_paths = list(filter(lambda p: p.exists(), paths))
    if not config_paths:
        raise ConfigurationError("No config file found.")

    cfg = configparser.ConfigParser()

    for path in config_paths:
        cfg.read(path)

    _update_config(cfg, opts)

    return cfg



"""
Store tasks in PostgreSQL.
"""
import logging
from typing import Iterable, Optional

import psycopg2 as pg

from taskrabbit.config import PostgresConfig
from .base import StoredTask, TaskStore


class PostgresTaskStore(TaskStore):
    config_class = PostgresConfig

    def __init__(self, cfg: PostgresConfig):
        self.conn = pg.connect(dsn=cfg.get_dsn())
        self.create_table()

    def create_table(self):
        with self.conn.cursor() as c:
            c.execute(
                """
            CREATE TABLE IF NOT EXISTS tasks (
                id text UNIQUE NOT NULL
                , task text
                , args text
                , kwargs text
                , task_data json
            )
            """
            )

    def execute(self, query: str, *params) -> pg.extensions.cursor:
        c = self.conn.cursor()
        try:
            logging.debug(c.mogrify(query, params))
            c.execute(query, params)
            self.conn.commit()
        except Exception as e:
            logging.exception(e)
            self.conn.rollback()
            raise
        return c

    def save(self, task: StoredTask):
        self.execute(
            """
            INSERT INTO tasks
            VALUES
                (%s, %s, %s, %s, %s)
            ON CONFLICT (id) DO NOTHING
        """,
            task.id,
            task.task,
            task.argsrepr,
            task.kwargsrepr,
            task.json(),
        )

    def bulk_save(self, tasks: Iterable[StoredTask]):
        # https://stackoverflow.com/a/12207237/21245
        # https://www.psycopg.org/docs/usage.html#using-copy-to-and-copy-from
        pass

    def delete(self, task: StoredTask):
        self.execute(
            """
        DELETE FROM tasks
        WHERE id=%s
        """,
            task.id,
        )

    def load_tasks(self, task_name: Optional[str] = None) -> Iterable[StoredTask]:
        if task_name is None:
            logging.debug("loading tasks")
            cursor = self.execute("SELECT task_data FROM tasks")
        else:
            logging.debug("loading tasks with task name: %s", task_name)
            cursor = self.execute(
                "SELECT task_data FROM tasks WHERE task=%s", task_name
            )
        for row in cursor.fetchall():
            logging.debug(f"{row=}")
            yield StoredTask(**row[0])

    def dedupe(self) -> int:
        cur = self.execute(
            """
        DELETE FROM tasks
        WHERE id not in (
            SELECT max(id) from tasks
            GROUP BY task, args, kwargs
        );
        """
        )
        return cur.rowcount



import logging
import socket
from collections import Counter
from itertools import islice
from typing import List, Optional, Iterable
import termtables
from kombu import Connection, Exchange, Message, Queue
from amqp.exceptions import NotFound as AMQPNotFound

from . import config
from .stores.base import StoredTask, TaskStore


class TaskCounter(Counter):
    def stream(self, tasks: Iterable[StoredTask]):
        for task in tasks:
            self.update([task.task])
            yield task

    def display(self):
        if self:
            termtables.print(self.most_common(), header=["Task", "Count"])


def fill(
    cfg: config.Config,
    exchange_name: str,
    store: TaskStore,
    task_name: Optional[str] = None,
    delete: bool = True,
) -> None:
    # Don't publish to system exchanges
    if exchange_name.startswith("amq"):
        raise ValueError(f"Cannot publish to system exchange: {exchange_name}")

    with Connection(cfg.rabbitmq.url()) as conn:
        with conn.channel() as channel:
            try:
                counter = TaskCounter()
                # Passively declare the exchange so we can fail if it doesn't
                # already exist.
                exchange = Exchange(exchange_name, channel=channel, passive=True)
                logging.debug("Established connection")
                producer = conn.Producer(
                    exchange=exchange, channel=channel, serializer="json"
                )
                for task in counter.stream(store.load_tasks(task_name)):
                    logging.debug("Publishing task ID: %s", task.id)
                    producer.publish(
                        task.body,
                        exchange=exchange,
                        routing_key=task.routing_key,
                        headers=task.headers,
                    )
                    if delete:
                        store.delete(task)
                counter.display()
            except AMQPNotFound as ex:
                logging.error(str(ex))


def drain(cfg: config.Config, queue_name: str, store: TaskStore) -> None:
    queue = Queue(queue_name)
    logging.info(f"Draining queue: {queue}")

    requeue: List[Message] = []

    def callback(_, message: Message):
        nonlocal requeue
        task = StoredTask.from_message(message)
        logging.debug("Received task: %s", task)
        try:
            store.save(task)
            message.ack()
        except Exception:
            requeue.append(message)

    # TODO: Safer persistence for messages that need to be requeued.
    with Connection(cfg.rabbitmq.url()) as conn:
        try:
            with conn.Consumer(queue, callbacks=[callback]) as consumer:
                consumer.qos(
                    prefetch_count=config.RabbitMQConfig.consumer_prefetch_count
                )
                try:
                    while True:
                        conn.drain_events(timeout=1)
                except socket.timeout:
                    pass
                except KeyboardInterrupt:
                    consumer.recover(requeue=True)
                    raise
        finally:
            if requeue:
                for message in requeue:
                    message.requeue()


def list_(
    store: TaskStore,
    counts=False,
    limit: Optional[int] = None,
    task_name: Optional[str] = None,
) -> None:
    stream = store.load_tasks(task_name)
    if limit:
        stream = islice(stream, limit)
    if counts:
        counter = TaskCounter()
        list(counter.stream(stream))
        counter.display()
    else:
        items = []
        for task in stream:
            items.append(
                (task.id, task.task, task.argsrepr, task.kwargsrepr, task.routing_key)
            )
        if items:
            termtables.print(
                items, header=["ID", "Task", "Args", "Kwargs", "Routing Key"]
            )



import logging
import sqlite3
from typing import Iterable, Optional

from taskrabbit.config import SqliteConfig
from .base import TaskStore, StoredTask


class SqliteTaskStore(TaskStore):
    config_class = SqliteConfig

    def __init__(self, cfg: SqliteConfig):
        super().__init__()
        self.conn = sqlite3.connect(cfg.db)
        self.conn.set_trace_callback(logging.debug)
        self.conn.row_factory = sqlite3.Row
        self.create_table()

        # execute many writes 1000x faster by not waiting for
        # each transaction to go to disk.
        # See https://www.sqlite.org/faq.html#q19
        self.execute("PRAGMA synchronous=OFF")

    def create_table(self):
        self.execute(
            """
        CREATE TABLE IF NOT EXISTS tasks (
            id text UNIQUE NOT NULL
            , task text
            , args text
            , kwargs text
            , json text
        )
        """
        )

    def execute(self, query: str, *params) -> sqlite3.Cursor:
        c = self.conn.cursor()
        try:
            c.execute(query, params)
            self.conn.commit()
        except Exception as e:
            logging.exception(e)
            raise
        return c

    def save(self, task: StoredTask):
        values = (
            task.id,
            task.task,
            task.argsrepr,
            task.kwargsrepr,
            task.json(indent=0),
        )
        self.execute(
            """
        INSERT OR IGNORE INTO tasks
        VALUES
            (?, ?, ?, ?, ?)
        """,
            *values
        )

    def delete(self, task: StoredTask):
        self.execute(
            """
        DELETE FROM tasks
        WHERE id=?""",
            task.id,
        )

    def load_tasks(self, task_name: Optional[str] = None) -> Iterable[StoredTask]:
        if task_name is None:
            logging.debug("loading tasks")
            cursor = self.execute("SELECT * FROM tasks")
        else:
            logging.debug("loading tasks with task name: '%s'", task_name)
            cursor = self.execute("SELECT * FROM tasks WHERE task=?", task_name)
        for row in cursor.fetchall():
            yield StoredTask.from_string(row["json"])

    def dedupe(self) -> int:
        cur = self.execute(
            """
        DELETE FROM tasks
        WHERE id not in (
            SELECT max(id) from tasks
            GROUP BY task, args, kwargs
        );
        """
        )
        return cur.rowcount



import logging
from enum import Enum
from pathlib import Path
from typing import Optional

import typer

from taskrabbit import __version__

from .config import (
    Config,
    ConfigurationError,
    merge_config_files_and_options,
)
from .operations import drain, fill, list_
from .utils import pluralize, green, red


HOME_CONFIG_PATH = Path.home() / ".taskrabbit.ini"

app = typer.Typer()
store_app = typer.Typer(short_help="Interact with stored tasks.")
app.add_typer(store_app, name="store")


class LogLevels(str, Enum):
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"
    critical = "critical"


@store_app.command("dedupe")
def de_duplicate(ctx: typer.Context):
    """
    Remove duplicate tasks.
    """
    confirmed = typer.confirm(red("This is a destructive operation. Continue?"))
    if not confirmed:
        raise typer.Abort()
    cfg = ctx.meta["config"]
    store = cfg.init_store()
    try:
        count = store.dedupe()
    except NotImplementedError:
        typer.echo(
            red(f"{store.__class__.__name__} does not support task de-duplication."),
        )
        raise typer.Exit(code=1)
    # count_display = typer.style(str(count), fg=typer.colors.GREEN)
    typer.echo(
        f"Removed {green(count)} duplicate task{pluralize(count)} from the store."
    )


@app.command("drain")
def drain_command(
    ctx: typer.Context,
    queue: str = typer.Argument(..., help="Queue to drain tasks from."),
) -> None:
    """
    Drain tasks from the queue.
    """
    cfg = ctx.meta["config"]
    store = cfg.init_store()
    drain(cfg, queue, store)
    print("Stored tasks:")
    list_(store, counts=True)


@app.command("fill")
def fill_command(
    ctx: typer.Context,
    exchange: str = typer.Argument("", help="Tasks will be published to this exchange"),
    task_name: Optional[str] = typer.Option(
        None, help="Only publish tasks with this name"
    ),
    delete: bool = typer.Option(True, help="Delete tasks from store after publishing."),
    confirm: bool = typer.Option(True, help="Confirm exchange before publishing"),
) -> None:
    """
    Publish tasks to an exchange.
    """

    # Confirm exhange
    exchange_display = exchange if exchange else "default"
    if confirm:
        confirmed = typer.confirm(f"Publish tasks to the {exchange_display} exchange?")
        if not confirmed:
            raise typer.Abort()
    cfg = ctx.meta["config"]
    store = cfg.init_store()
    fill(cfg, exchange, store, task_name, delete)


@store_app.command("list")
def list_command(
    ctx: typer.Context,
    counts: bool = typer.Option(
        False, "--counts", "-c", help="Only show task counts, by name"
    ),
    task: Optional[str] = typer.Option(None, help="List only tasks with this name."),
    limit: Optional[int] = typer.Option(None, help="Limit number of tasks shown"),
) -> None:
    """
    Show retrieved tasks.
    """
    cfg = ctx.meta["config"]
    store = cfg.init_store()
    list_(store, counts=counts, task_name=task, limit=limit)


def show_version(value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()


def check_config(value) -> Optional[Path]:
    if value is None:
        path = Path("taskrabbit.ini")
        if path.exists():
            return path
    else:
        path = value.absolute()
        if not path.exists():
            raise typer.BadParameter(f"Config file does not exist at {path}")
        return value
    return None


@app.callback()
def main(
    ctx: typer.Context,
    config: Optional[Path] = typer.Option(
        None,
        "--config",
        "-c",
        exists=True,
        readable=True,
        resolve_path=True,
        callback=check_config,
    ),
    log_level: str = typer.Option(LogLevels.info, case_sensitive=False),
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=show_version,
        help="Display the program version.",
        is_eager=True,
    ),
):
    """
    Remove and restore Celery tasks from RabbitMQ.
    """
    config_paths = [HOME_CONFIG_PATH]
    if config is not None:
        config_paths.append(config)
    try:
        cfg_dict = merge_config_files_and_options(
            *config_paths, taskrabbit={"log_level": log_level.upper()}
        )
        cfg = Config.from_config_dict(cfg_dict)
        log_level = getattr(logging, cfg.log_level)
        logging.basicConfig(level=log_level)
        logging.debug("Loaded configuration: %s", cfg)

        ctx.meta["config"] = cfg
    except ConfigurationError as exc:
        raise typer.BadParameter(str(exc)) from exc



from setuptools import setup
import os
import sys
from pathlib import Path


assert sys.version_info >= (3, 6, 0), "taskrabbit requires Python 3.6+"

PROJECT_DIR = Path(__file__).parent
sys.path.insert(0, str(PROJECT_DIR))

EXTENSIONS = ["postgres"]


def get_long_description() -> str:
    return (PROJECT_DIR / "README.md").read_text(encoding="utf8")


# REQUIREMENTS
# requirements parsing borrowed from Celery project.


def _strip_comments(line):
    return line.split("#", 1)[0].strip()


def _pip_requirement(req):
    if req.startswith("-r "):
        _, path = req.split()
        return reqs(*path.split("/"))
    return [req]


def _reqs(*f):
    return [
        _pip_requirement(r)
        for r in (
            _strip_comments(line)
            for line in open(os.path.join(os.getcwd(), "requirements", *f)).readlines()
        )
        if r
    ]


def reqs(*f):
    """Parse requirement file.
    Example:
        reqs('default.txt')          # requirements/default.txt
        reqs('extras', 'redis.txt')  # requirements/extras/redis.txt
    Returns:
        List[str]: list of requirements specified in the file.
    """
    return [req for subreq in _reqs(*f) for req in subreq]


def extras(*p):
    """Parse requirement in the requirements/extras/ directory."""
    return reqs("extras", *p)


def install_requires():
    """Get list of requirements required for installation."""
    return reqs("base.txt")


def extras_require():
    """Get map of all extra requirements."""
    return {x: extras(x + ".txt") for x in EXTENSIONS}


setup(
    name="taskrabbit",
    version="0.4.1",
    description="Move Celery tasks in and out of RabbitMQ",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Chris Lawlor",
    url="https://github.com/chrislawlor/taskrabbit",
    license="MIT",
    packages=["taskrabbit", "taskrabbit.stores"],
    python_requires=">=3.6",
    zip_safe=False,
    install_requires=install_requires(),
    extras_require=extras_require(),
    entry_points={
        "console_scripts": [
            "taskr = taskrabbit.__main__:main",
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: System :: Distributed Computing",
        "Topic :: Software Development :: Object Brokering",
        "Operating System :: OS Independent",
    ],
)
